@@ -18,6 +18,7 @@ class GA:
                  K_tournament=3,
                  crossover_type="single_point",
                  mutation_type="random",
+                 mutation_by_replacement=False,
                  mutation_percent_genes=10,
                  mutation_num_genes=None,
                  random_mutation_min_val=-1.0,
@@ -44,8 +45,10 @@ class GA:
         keep_parents: If 0, this means the parents of the current populaiton will not be used at all in the next population. If -1, this means all parents in the current population will be used in the next population. If set to a value > 0, then the specified value refers to the number of parents in the current population to be used in the next population. In some cases, the parents are of high quality and thus we do not want to loose such some high quality solutions. If some parent selection operators like roulette wheel selection (RWS), the parents may not be of high quality and thus keeping the parents might degarde the quality of the population.
         K_tournament: When the value of 'parent_selection_type' is 'tournament', the 'K_tournament' parameter specifies the number of solutions from which a parent is selected randomly.
 
-        crossover_type: Type of the crossover opreator.
-        mutation_type: Type of the mutation opreator.
+        crossover_type: Type of the crossover opreator. If  crossover_type=None, then the crossover step is bypassed which means no crossover is applied and thus no offspring will be created in the next generations. The next generation will use the solutions in the current population.
+        mutation_type: Type of the mutation opreator. If mutation_type=None, then the mutation step is bypassed which means no mutation is applied and thus no changes are applied to the offspring created using the crossover operation. The offspring will be used unchanged in the next generation.
+
+        mutation_by_replacement: An optional bool parameter. It works only when the selected type of mutation is random (mutation_type="random"). In this case, setting mutation_by_replacement=True means replace the gene by the randomly generated value. If False, then it has no effect and random mutation works by adding the random value to the gene.
 
         mutation_percent_genes: Percentage of genes to mutate which defaults to 10%. This parameter has no action if the parameter mutation_num_genes exists.
         mutation_num_genes: Number of genes to mutate which defaults to None. If the parameter mutation_num_genes exists, then no need for the parameter mutation_percent_genes.
@@ -98,20 +101,6 @@ class GA:
 
         self.num_parents_mating = num_parents_mating
 
-        if (mutation_num_genes == None):
-            if (mutation_percent_genes < 0 or mutation_percent_genes > 100):
-                self.valid_parameters = False
-                raise ValueError("The percentage of selected genes for mutation (mutation_percent_genes) must be >= 0 and <= 100 inclusive but {mutation_percent_genes=mutation_percent_genes} found.\n".format(mutation_percent_genes=mutation_percent_genes))
-        elif (mutation_num_genes <= 0 ):
-            self.valid_parameters = False
-            raise ValueError("The number of selected genes for mutation (mutation_num_genes) cannot be <= 0 but {mutation_num_genes} found.\n".format(mutation_num_genes=mutation_num_genes))
-        elif (mutation_num_genes > self.num_genes):
-            self.valid_parameters = False
-            raise ValueError("The number of selected genes for mutation (mutation_num_genes) ({mutation_num_genes}) cannot be greater than the number of genes ({num_genes}).\n".format(mutation_num_genes=mutation_num_genes, num_genes=self.num_genes))
-        elif (type(mutation_num_genes) is not int):
-            self.valid_parameters = False
-            raise ValueError("The number of selected genes for mutation (mutation_num_genes) must be a positive integer >= 1 but {mutation_num_genes} found.\n".format(mutation_num_genes=mutation_num_genes))
-
         # crossover: Refers to the method that applies the crossover operator based on the selected type of crossover in the crossover_type property.
         # Validating the crossover type: crossover_type
         if (crossover_type == "single_point"):
@@ -120,9 +109,11 @@ class GA:
             self.crossover = self.two_points_crossover
         elif (crossover_type == "uniform"):
             self.crossover = self.uniform_crossover
+        elif (crossover_type is None):
+            self.crossover = None
         else:
             self.valid_parameters = False
-            raise ValueError("ERROR: undefined crossover type. \nThe assigned value to the crossover_type ({crossover_type}) argument does not refer to one of the supported crossover types which are: \n-single_point (for single point crossover)\n-two_points (for two points crossover)\n-uniform (for uniform crossover).\n".format(crossover_type=crossover_type))
+            raise ValueError("Undefined crossover type. \nThe assigned value to the crossover_type ({crossover_type}) argument does not refer to one of the supported crossover types which are: \n-single_point (for single point crossover)\n-two_points (for two points crossover)\n-uniform (for uniform crossover).\n".format(crossover_type=crossover_type))
 
         self.crossover_type = crossover_type
 
@@ -136,12 +127,43 @@ class GA:
             self.mutation = self.scramble_mutation
         elif (mutation_type == "inversion"):
             self.mutation = self.inversion_mutation
+        elif (mutation_type is None):
+            self.mutation = None
         else:
             self.valid_parameters = False
-            raise ValueError("ERROR: undefined mutation type. \nThe assigned value to the mutation_type argument does not refer to one of the supported mutation types which are: \n-random (for random mutation)\n-swap (for swap mutation)\n-inversion (for inversion mutation)\n-scramble (for scramble mutation).\n")
+            raise ValueError("Undefined mutation type. \nThe assigned value to the mutation_type argument ({mutation_type}) does not refer to one of the supported mutation types which are: \n-random (for random mutation)\n-swap (for swap mutation)\n-inversion (for inversion mutation)\n-scramble (for scramble mutation).\n".format(mutation_type=mutation_type))
 
         self.mutation_type = mutation_type
 
+        if not (self.mutation_type is None):
+            if (mutation_num_genes == None):
+                if (mutation_percent_genes < 0 or mutation_percent_genes > 100):
+                    self.valid_parameters = False
+                    raise ValueError("The percentage of selected genes for mutation (mutation_percent_genes) must be >= 0 and <= 100 inclusive but {mutation_percent_genes=mutation_percent_genes} found.\n".format(mutation_percent_genes=mutation_percent_genes))
+            elif (mutation_num_genes <= 0 ):
+                self.valid_parameters = False
+                raise ValueError("The number of selected genes for mutation (mutation_num_genes) cannot be <= 0 but {mutation_num_genes} found.\n".format(mutation_num_genes=mutation_num_genes))
+            elif (mutation_num_genes > self.num_genes):
+                self.valid_parameters = False
+                raise ValueError("The number of selected genes for mutation (mutation_num_genes) ({mutation_num_genes}) cannot be greater than the number of genes ({num_genes}).\n".format(mutation_num_genes=mutation_num_genes, num_genes=self.num_genes))
+            elif (type(mutation_num_genes) is not int):
+                self.valid_parameters = False
+                raise ValueError("The number of selected genes for mutation (mutation_num_genes) must be a positive integer >= 1 but {mutation_num_genes} found.\n".format(mutation_num_genes=mutation_num_genes))
+        else:
+            pass
+
+        if not (type(mutation_by_replacement) is bool):
+            self.valid_parameters = False
+            raise TypeError("The expected type of the 'mutation_by_replacement' parameter is bool but {mutation_by_replacement_type} found.".format(mutation_by_replacement_type=type(mutation_by_replacement)))
+
+        self.mutation_by_replacement = mutation_by_replacement
+        
+        if self.mutation_type != "random" and self.mutation_by_replacement:
+            print("Warning: The mutation_by_replacement parameter is set to True while the mutation_type parameter is not set to random but {mut_type}. Note that the mutation_by_replacement parameter has an effect only when mutation_type='random'.".format(mut_type=mutation_type))
+
+        if (self.mutation_type is None) and (self.crossover_type is None):
+            print("Warning: the 2 parameters mutation_type and crossover_type are None. This disables any type of evolution the genetic algorithm can make. As a result, the genetic algorithm cannot find a better solution that the best solution in the initial population.")
+
         # select_parents: Refers to a method that selects the parents based on the parent selection type specified in the parent_selection_type attribute.
         # Validating the selected type of parent selection: parent_selection_type
         if (parent_selection_type == "sss"):
@@ -213,7 +235,7 @@ class GA:
             self.callback_generation = None
 
         # The number of completed generations.
-        self.generations_completed = None 
+        self.generations_completed = 0
 
         # At this point, all necessary parameters validation is done successfully and we are sure that the parameters are valid.
         self.valid_parameters = True # Set to True when all the parameters passed in the GA class constructor are valid.
@@ -255,7 +277,7 @@ class GA:
                                                size=self.pop_size) # A NumPy array holding the initial population.
         
         # Keeping the initial population in the initial_population attribute.
-        self.initial_population = self.population
+        self.initial_population = self.population.copy()
 
     def cal_pop_fitness(self):
 
@@ -297,12 +319,23 @@ class GA:
             # Selecting the best parents in the population for mating.
             parents = self.select_parents(fitness, num_parents=self.num_parents_mating)
 
-            # Generating next generation using crossover.
-            offspring_crossover = self.crossover(parents,
-                                                 offspring_size=(self.num_offspring, self.num_genes))
+            # If self.crossover_type=None, then no crossover is applied and thus no offspring will be created in the next generations. The next generation will use the solutions in the current population.
+            if self.crossover_type is None:
+                if self.num_offspring <= self.keep_parents:
+                    offspring_crossover = parents[0:self.num_offspring]
+                else:
+                    offspring_crossover = numpy.concatenate((parents, self.population[0:(self.num_offspring - parents.shape[0])]))
+            else:
+                # Generating offspring using crossover.
+                offspring_crossover = self.crossover(parents,
+                                                     offspring_size=(self.num_offspring, self.num_genes))
 
-            # Adding some variations to the offspring using mutation.
-            offspring_mutation = self.mutation(offspring_crossover)
+            # If self.mutation_type=None, then no mutation is applied and thus no changes are applied to the offspring created using the crossover operation. The offspring will be used unchanged in the next generation.
+            if self.mutation_type is None:
+                offspring_mutation = offspring_crossover
+            else:
+                # Adding some variations to the offspring using mutation.
+                offspring_mutation = self.mutation(offspring_crossover)
 
             if (self.keep_parents == 0):
                 self.population = offspring_mutation
@@ -572,9 +605,14 @@ class GA:
         # Random mutation changes a single gene in each offspring randomly.
         for offspring_idx in range(offspring.shape[0]):
             for gene_idx in mutation_indices:
-                # The random value to be added to the gene.
+                # Generating a random value.
                 random_value = numpy.random.uniform(self.random_mutation_min_val, self.random_mutation_max_val, 1)
-                offspring[offspring_idx, gene_idx] = offspring[offspring_idx, gene_idx] + random_value
+                # If the mutation_by_replacement attribute is True, then the random value replaces the current gene value.
+                if self.mutation_by_replacement:
+                    offspring[offspring_idx, gene_idx] = random_value
+                # If the mutation_by_replacement attribute is False, then the random value is added to the gene value.
+                else:
+                    offspring[offspring_idx, gene_idx] = offspring[offspring_idx, gene_idx] + random_value
         return offspring
 
     def swap_mutation(self, offspring):
